/* ============================================================================================================================================= /
 * PROJECT:         Orologio 4digig
 * FILE:            Main.c
 * LAST REV:        1.2
 * AUTHOR:          Ivo Colleoni
 * DEVICE:          PIC18F2550
 * CREATED:         15/07/2014
 * UPDATED:
 *
 * DESCRIP:         Orologio digitale su display 4 cifre, scrittura in multiplex continuo, orologio tramite timer e clock esterno a 32.768KHz
 *
 * ============================================================================================================================================= /
 * Change History:
 * Rev.             Date                Changes
 * 1.2              15/07/2014          - Aggiunto Delay da FINE_TUNE ogni secondo per ridurre l'errore di -1s ogni 24h
 * 1.1              06/07/2014          - Prima release
 * 1.0              25/06/2014          - Inizio sviluppo
/* ============================================================================================================================================ */

/* INCLUDES =================================================================================================================================== */
#include <xc.h>
#include "Main.h"
#include "Prototipi.h"
#include "Delay.c"
#include "Manage_input.c"

/* GLOBAL VARIABLES =========================================================================================================================== */
unsigned char hours = 0;                                                                // ore in PBCD
unsigned char minutes = 0;                                                              // minuti in PBCD
unsigned char seconds = 0;                                                              // secondi in PBCD
unsigned char persist = 0;
unsigned char counter  = 0;

bit dots = 0;
bit setted = 0;
bit show_hours = 0;
bit show_minutes = 0;
bit blink = 0;
bit global_blink = 0;

unsigned char decode_table[10] = {
                         0b00111111,    // 0
                         0b00000110,    // 1
                         0b01011011,    // 2
                         0b01001111,    // 3
                         0b01100110,    // 4
                         0b01101101,    // 5
                         0b01111101,    // 6
                         0b00000111,    // 7
                         0b01111111,    // 8
                         0b01101111     // 9
};

/* ============================================================================================================================================= /
 * FUNCTION:       Interrupts
 *
 * Overview:
/* ============================================================================================================================================ */
void interrupt int_high (void) {

    if (PIR1bits.TMR1IF) {                                                              // interrupt causato da overflow timer 1 (ogni 1 secondi)
        TMR1 = TMR1_PRELOAD;                                                            // preload timer 1   (prima di altre operazioni per avere interrupt costanti)

        dots = !dots;                                                                   // inverto lo stato dei puntini

        seconds = Inc_pbcd (seconds, MIN_SEC, MAX_SEC);                                 // incremento i secondi
        T1CONbits.TMR1ON = 0;                                                           // fermo timer 1
        Delay_us(FINE_TUNE);                                                            // aggiungo delay
        T1CONbits.TMR1ON = 1;                                                           // riavvio timer 1

        if (seconds == 0) {                                                             // se i secondi dopo l'incremento sono 0 (59+1)
            minutes = Inc_pbcd (minutes, MIN_MIN, MAX_MIN);                             // incremento i minuti

            if (minutes == 0){                                                          // se i minuti dopo l'incremento sono 0
                hours = Inc_pbcd (hours, MIN_HOU, MAX_HOU);                             // incremento le ore
            }
        }

        PIR1bits.TMR1IF = 0;
    }
}

void interrupt low_priority int_low(void) {

    if (PIR1bits.TMR2IF) {                                                              // interrupt causato da timer 2 (1ms)
        static unsigned char _timer_multiplex = 0;
        static unsigned int _timer_blink = 0;
        unsigned char _value = 0;

        _timer_multiplex++;
        _timer_blink ++;                                                                // incremento contatori

        if (_timer_multiplex == DISPLAY_MUX) {                                          // se contatore arriva a DISPLAY_MUX (1ms * DISPLAY_MUX)
                                                                                        // aggiorno i display
            if (show_hours){                                                            // se devo esporre le ore
                _value = (hours & 0b11110000);                                          // Estraggo valore decine ore
                _value = _value >> 4;
                Cathode(0);                                                             // spengo tutti i catodi
                DISPLAY = decode_table[_value];                                         // decodifico il valore da mandare al display
                Cathode(1);                                                             // abilito catodo decine ore
                Delay_us(PERSIST);

                _value = (hours & 0b00001111);                                          // Estraggo valore ore
                Cathode(0);
                DISPLAY = decode_table[_value];
                
                DP = dots;                                                              // accendo o spengo il putnino
                Cathode(2);
                Delay_us(PERSIST);                                                      // ritardo permanenza valore su display
                Cathode(0);                                                             // spengo tutti i catodi
            }

            if (show_minutes){                                                          // se devo esporre i minuti
                _value = (minutes & 0b11110000);                                        // Estraggo valore decine minuti
                _value = _value >> 4;
                Cathode(0);
                DISPLAY = decode_table[_value];
                DP = dots;
                Cathode(3);
                Delay_us(PERSIST);

                _value = (minutes & 0b00001111);                                        // Estraggo valore minuti
                Cathode(0);
                DISPLAY = decode_table[_value];
                Cathode(4);
                Delay_us(PERSIST);
                Cathode(0);
            }
            _timer_multiplex = 0;                                                       // resetto contatore timer
        }

        if (_timer_blink == BLINK_DELAY) {                                              // 1ms * BLINK_DELAY
            blink = !blink;                                                             // inverto flang blink
            _timer_blink = 0;
        }

        TMR2 = 131;
        PIR1bits.TMR2IF = 0;
    }
}

/* ============================================================================================================================================= /
 * FUNCTION:        Main
 *
 * Overview:        se pulsante SET_SW  premuto va al setup altrimenti scrive costantemente i display. Se non  mai stato impostato lampeggia
/* ============================================================================================================================================ */

void main(void) {
    Define_IO();
    Module_settings();

    show_hours = 1;                                                                     // valori standard
    show_minutes = 1;
    Cathode(0);
    DISPLAY = 0;

    Setup();                                                                            // al primo avvio vado subito al setup

    while (1) {
        if (Debounce_single(SET_SW))                                                    // se setup  premuto vado al setup
            Setup();

    }
}

/* ============================================================================================================================================= /
 * FUNCTION:        Setup
 *
 * Overview:        Imposta prima i minuti e poi le ore ad ogni pressione di SET_SW
/* ============================================================================================================================================ */

void Setup(void){
    unsigned char    _previus_value = 0;
    PIE1bits.TMR1IE = 0;                                                                // disabilito interrupt timer 1
    dots = ON;                                                                          // forzo accensione puntini

    if (!setted) {                                                                      // al primo avvio lampeggio sia ore che mninuti
        do {
            if (blink) {
                show_hours = 0;
                show_minutes = 0;
            } else {
                show_hours = 1;
                show_minutes = 1;
            }
        }   while (!Debounce_single(SET_SW) && !Debounce_single(INC_SW));               // fino a quando non viene premuto un tasto
    }

    while (Debounce_single(INC_SW))                                                     // attendo rilascio inc sw
        continue;

    _previus_value = hours;                                                             // salvo il valore delle ore in una variabile temporanea

    do {
        if (Debounce_multi(INC_SW)) {                                                   // se pulsante incremento  premuto
            hours = Inc_pbcd(hours, MIN_HOU, MAX_HOU);                                  // incremeto il byte dei minuti
        }

        if (_previus_value == hours)                                                    // se il valore delle ore non  cambiato
            global_blink = 1;                                                           // consenti il lampeggio
        else {                                                                          // altrimenti non lampeggiare
            global_blink = 0;
            _previus_value = hours;
        }

        if (blink && global_blink) {
            show_hours = 0;
            show_minutes = 1;
        } else {
            show_hours = 1;
            show_minutes = 1;
        }
    } while (!Debounce_single(SET_SW));                                                 // se viene premuto SET_SW proseguo

    _previus_value = minutes;

    do {
        if (Debounce_multi(INC_SW)) {                                                   // se pulsante incremento  premuto
            minutes = Inc_pbcd(minutes, MIN_MIN, MAX_MIN);                              // incremeto il byte dei minuti
        }

        if (_previus_value == minutes)
            global_blink = 1;
        else {
            global_blink = 0;
            _previus_value = minutes;
        }

        if (blink && global_blink) {
            show_hours = 1;
            show_minutes = 0;
        } else {
            show_hours = 1;
            show_minutes = 1;
        }
    } while (!Debounce_single(SET_SW));                                                 // se viene premuto SET_SW proseguo


    blink = 0;
    show_hours = 1;
    show_minutes = 1;


    seconds = 0;                                                                        // resetto i secondi
    setted = 1;                                                                         // setto il flag di setup avvenuto

    TMR1 = TMR1_PRELOAD;                                                                // resetto preload timer 1
    PIE1bits.TMR1IE = 1;                                                                // abilito interrupt timer 1
}

/* ============================================================================================================================================= /
 * FUNCTION:    Inc_pbcd
 * Input:       _value: valore attuale da incrementare
 *              _min: valore minimo consentito del risultato
 *              _max: valore massimo consentito del risultato
 * Output:      Ritorna il valore incrementato di 1 unit
 *
 * Overview:    Incrementa di 1 unit il singolo byte (unsigned char) in packed BCD (4bit alti: decine, 4 bit bassi: unit (max 9) mantenendo il risultato
 *              tra il valore minimo e il valore massimo consentito
/* ============================================================================================================================================ */

unsigned char Inc_pbcd(unsigned char _value, unsigned char _min, unsigned char _max) {
    _value++;                                                                           // incremento il valore
    if ((_value & 0b00001111) == 10)
        _value = _value + 6;                                                            // se risultato degli ultimi 4 bit  10 sommo 6 al risultato
    if (_value > _max)
        _value = _min;                                                                  // se risultato supera il valore massimo consentito riporto
                                                                                        // il risultato al valore minimo consentito
    return(_value);
}
/* ============================================================================================================================================= /
 * FUNCTION:
 *
 * Overview:    Aggiorna i valori su display
/* ============================================================================================================================================ */
void Cathode(unsigned char _switch) {
    switch(_switch) {
        case 0: CATHODE_1 = OFF;
                CATHODE_2 = OFF;
                CATHODE_3 = OFF;
                CATHODE_4 = OFF;
                break;
        case 1: CATHODE_1 = ON;
                CATHODE_2 = OFF;
                CATHODE_3 = OFF;
                CATHODE_4 = OFF;
                break;
        case 2: CATHODE_1 = OFF;
                CATHODE_2 = ON;
                CATHODE_3 = OFF;
                CATHODE_4 = OFF;
                break;
        case 3: CATHODE_1 = OFF;
                CATHODE_2 = OFF;
                CATHODE_3 = ON;
                CATHODE_4 = OFF;
                break;
        case 4: CATHODE_1 = OFF;
                CATHODE_2 = OFF;
                CATHODE_3 = OFF;
                CATHODE_4 = ON;
                break;
        case 5: CATHODE_1 = ON;
                CATHODE_2 = ON;
                CATHODE_3 = ON;
                CATHODE_4 = ON;
                break;
        }
}




